#include <stdio.h>
#include <stdint.h>
#include <math.h>

#if X86
static inline uint64_t get_cycles()
{
  uint64_t t;
  __asm volatile ("rdtsc" : "=A"(t));
  return t;
}
#endif

#if ARM
// This instruction must be enabled in kernel mode for example by loading a module which will enable it.
static unsigned int get_cycles(void)
{
    unsigned int value;

    /* Read CCNT Register */
    asm volatile ("mrc p15, 0, %0, c9, c13, 0\t\n": "=r"(value));

    return value;
}

static void init_perfcounters(int32_t do_reset, int32_t enable_divider)
{
    /* In general enable all counters (including cycle counter) */
    int32_t value = 1;

    /* Peform reset */
    if (do_reset) {
        value |= 2; /* reset all counters to zero */
        value |= 4; /* reset cycle counter to zero */
    }

    if (enable_divider)
        value |= 8; /* enable "by 64" divider for CCNT */

    value |= 16;

    /* Program the performance-counter control-register */
    asm volatile ("mcr p15, 0, %0, c9, c12, 0\t\n" :: "r"(value));

    /* Enable all counters */
    asm volatile ("mcr p15, 0, %0, c9, c12, 1\t\n" :: "r"(0x8000000f));

    /* Clear overflows */
    asm volatile ("mcr p15, 0, %0, c9, c12, 3\t\n" :: "r"(0x8000000f));
}
#endif

extern void FUNNAME (uint64_t *, uint64_t *);

int main () {
  uint64_t start, end;
  uint64_t ma[24], mc[24];
  int i, loop, i0, i1, R0, R1;
  double r[100], t, mean, var, std;

#if ARM
  init_perfcounters(1, 0);
#endif

  loop = 10*100;

  i1 = 100;
  t = 0;
  for (i0=0; i0 < i1; i0++) {
    start = get_cycles();
    for (i=0; i < loop/10; i++) {
      FUNNAME(ma, mc);
      FUNNAME(mc, ma);
      FUNNAME(ma, mc);
      FUNNAME(mc, ma);
      FUNNAME(ma, mc);
      FUNNAME(mc, ma);
      FUNNAME(ma, mc);
      FUNNAME(mc, ma);
      FUNNAME(ma, mc);
      FUNNAME(mc, ma);
    }
    end = get_cycles ();
    r[i0] = (double)(end-start)/loop;
    t += r[i0];
  }
  
  //for (i0=0; i0<i1; i0++) printf ("%d: %f\n", i0, r[i0]);

  mean = t/i1;
  var=0;
  for (i0=0; i0<i1; i0++) var += (mean-r[i0])*(mean-r[i0]);
  var /= i1;
  std = sqrt(var);

  printf ("%f cycles with std=%f\n", mean, std);

  R0 = i1;

  do {
  
    /* Remove everything which is more than 2 std away */
    for (i0=0; i0<i1; i0++) {
      if (r[i0] != -1 && r[i0] > (mean+2.5*std)) r[i0] = -1;
    }

    mean = 0; R1 = 0;
    for (i0=0; i0<i1; i0++) {
      if (r[i0] == -1) continue;
      mean += r[i0];
      R1++;
    }
    mean /= R1;

    var=0;
    for (i0=0; i0<i1; i0++) {
      if (r[i0] == -1) continue;
      var += (mean-r[i0])*(mean-r[i0]);
    }
    var /= R1;
    std = sqrt(var);

    printf ("From %d to %d\n", R0, R1);
    if (R0-R1 == 0) break;
    R0=R1;
    printf ("%f cycles with std=%f\n", mean, std);
  } while (0);


  return 0;
}

