#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <gmp.h>

// 352
#define BIT_LOW  (11*32)

// 480
#define BIT_HIGH (15*32)

#define NUMPRIMES 11
uint32_t primes[NUMPRIMES] = { 59, 17, 19, 23, 29, 31, 37, 41, 43, 47, 53 };

// stores primepowers
static mpz_t *cur;
// counts checked numbers
static uint32_t check = 0;

void search (int start, const int end) {
    // initialize all prime factors 1, ie the exponents to 0
    for (int j=0; j < NUMPRIMES; j++) {
        mpz_set_ui (cur[j], 1);
    }

    // initialize cur[0] to prime^start
    for (int j=0; j < start; j++) {
        mpz_mul_ui (cur[0], cur[0], primes[0]);
    }

    // initialize cur[j] to the value of cur[0]
    for (int j=1; j < NUMPRIMES; j++) {
        mpz_set (cur[j], cur[0]);
    }

    // temp storage for inside the loop
    mpz_t num;
    mpz_init (num);

    // this is the position at which we are currently counting
    // we start at the highest index
    int index = NUMPRIMES-1;

    while (1) {
        //for (int i=0; i<NUMPRIMES; i++) {
        //    printf("%2d: ", i);
        //    mpz_out_str (0, 10, cur[i]);
        //    printf ("\n");
        //}
        //printf("\n");

        int bitlength = mpz_sizeinbase (cur[index],2);

        // exceeded upper bound => go back to highest index
        if (bitlength >= BIT_HIGH) {
            index = NUMPRIMES-1;
        // exceeded lower index
        } else if (bitlength > BIT_LOW) {
            check++;
            // the real number is the one at the current index
            mpz_out_str (NULL, 10, cur[index]); printf ("\n");        
        }

        // increase number at index by the corresponding prime
        mpz_mul_ui (cur[index], cur[index], primes[index]);

        // exceeded upper bound, same check as above but now with one more
        if (bitlength >= BIT_HIGH) {
            while (bitlength >= BIT_HIGH) {
                // we have increased all numbers to the maximum => done
                if (index == 0) {
                    return;
                }

                // go to next index and increase it
                index--;

                // count how many times we have increased the highest prime
                // stop early if the start power is greater than the end power
                if (index == 0) {
                    start++;
                    if (start > end) {
                        return;
                    }
                }

                // increase current index by the appropriate prime
                mpz_mul_ui (cur[index], cur[index], primes[index]);
                // and update bitlength
                bitlength = mpz_sizeinbase (cur[index], 2);

            }

            // reset all lower positions to current highest prime power
            // effectively setting their power to 0
            for (int j=index+1; j < NUMPRIMES; j++) {
                mpz_set (cur[j], cur[index]);
            }
        }
        index = NUMPRIMES-1;
    }
}

int main (int argc, char *argv[]) {
    if(argc < 3) {
        printf("Arguments: START END\n");
        printf("START and END are powers of the highest prime\n");
        printf("both are inclusive\n");
        exit(1);
    }

    cur = (mpz_t *) malloc (NUMPRIMES*sizeof(mpz_t));
    for (int i=0; i < NUMPRIMES; i++) {
        mpz_init (cur[i]);
    }

    const int start = atoi (argv[1]);
    const int end = atoi (argv[2]);
    printf("end: %d\n", end);
    search (start, end);
    printf ("Check = %d\n", check);
    return 0;
}
